/***********************************************************
constants
***********************************************************/

// reference to the interface defined in cmguiIConnectedThresholdFilter.idl
const cmguiIConnectedThresholdFilter = Components.interfaces.cmguiIConnectedThresholdFilter;

// reference to the required base interface that all components must support
const nsISupports = Components.interfaces.nsISupports;

// references to other required interfaces
// const nsIDOMWindow = Components.interfaces.nsIDOMWindow;

// UUID uniquely identifying our component
// You can get from: http://kruithof.xs4all.nl/uuid/uuidgen here
const CLASS_ID = Components.ID("{E5ED99C8-CABD-11DB-898C-1B7156D89593}");

// description
const CLASS_NAME = "Cmgui connected threshold filter XPCOM Component";

// textual unique identifier
const CONTRACT_ID = "@bioeng.auckland.ac.nz/connectedthresholdfilter;1";

/***********************************************************
class definition
***********************************************************/

//class constructor
function ConnectedThresholdFilter() {
};


// class definition
ConnectedThresholdFilter.prototype = {

  // members
	parent: null,
	commandData: null,
	inputFieldName: null,
  inputField: null,
	outputFieldName: null,
  outputField: null,
	callback: null,

  belowValue: 0.4,
  aboveValue: 0.6,
  replaceValue: 1,
	numSeedPoints: 1,
	dimension: 2,
	seedValues: null,
	numSeedValues: 2,

	resampledInputFieldName: null,
	previewOutputFieldName: null,
	resampledInputField: null,
	previewOutputField: null,
	previewResolution: 128,
	gElementName: null,
	sceneViewer: null,


// define the function methods we want to expose in our interface

  init: function(parentWindow, commandData, inputFieldName, outputFieldName, callback) {
	 
		this.parent = parentWindow;
		this.commandData = commandData;
		this.inputFieldName = inputFieldName;
		this.outputFieldName = outputFieldName;
		this.callback = callback;

		// find the input field, note it is assumed to exist
	  this.inputField = commandData.findFieldByName(inputFieldName);

		// set some defaults which will be overwritten if the field exists
		this.belowValue = 0.4;
		this.aboveValue = 0.6;
		this.replaceValue = 1;
		this.numSeedPoints = 1;
		this.dimension = 2;
		this.seedValues = new Array();
		this.seedValues[0] = 0.5;
		this.seedValues[1] = 0.5;

		this.numSeedValues = 2;


		// check if output field exists
		if( commandData.isFieldDefined(outputFieldName) ){

			this.outputField = commandData.findFieldByName(outputFieldName);

			// create some objects which will be populated by the current field
			// properties if the field is the right type
			
			var returnInputField = new Object();
			var belowValue = new Object();
			var aboveValue = new Object();
			var replaceValue = new Object();
			var numSeedPoints = new Object();
			var dimension = new Object();
			var seedValues = new Object();
			var numSeedValues = new Object();
			
			try {

				this.outputField.getTypeConnectedThresholdImageFilter(returnInputField,belowValue,aboveValue,replaceValue,numSeedPoints,dimension,seedValues,numSeedValues);
				
				// when fetched the parameters are assigned to the value property

				this.belowValue = belowValue.value;
				this.aboveValue = aboveValue.value;
				this.replaceValue = replaceValue.value;
				this.numSeedPoints = numSeedPoints.value;
				this.dimension = dimension.value;
				this.seedValues = seedValues.value;
				this.numSeedValues = numSeedValues.value;
				
			} catch(ex) {

				this.parent.alert("Error fetching output field parameters.  The field exists but is probably of the wrong type:" + ex);

				this.outputField.setTypeConnectedThresholdImageFilter(this.inputField,this.belowValue,this.aboveValue,
          this.replaceValue,this.numSeedPoints,this.dimension,this.seedValues,this.numSeedValues);
				//				this.outputField.setTypeConnectedThresholdImageFilter(this.inputField,this.belowValue,this.aboveValue,this.replaceValue,this.xSeed,this.ySeed);

			}
			
		} else {

			// field does not exist so create it.

			this.outputField = commandData.createField(outputFieldName);
			this.outputField.setTypeConnectedThresholdImageFilter(this.inputField,this.belowValue,this.aboveValue,
  			this.replaceValue,this.numSeedPoints,this.dimension,this.seedValues,this.numSeedValues);
			
		}
    
  },


	// The connected threshold filter is very dependent on the resolution of the input image.
	// This means we can not resample our input field and then filter it to get a preview image.
	// Instead we must filter and then resample.
	setup_preview_field: function() {


 		// Set up previewOutputField
		this.previewOutputFieldName = "preview_" + this.outputFieldName;

		//		this.previewOutputField = this.commandData.createField(this.previewOutputFieldName);
		this.commandData.executeCommand("gfx def field " + this.previewOutputFieldName +
		 " image_resample field " + this.outputFieldName +
      " sizes " + this.previewResolution + " " + this.previewResolution + ";");
			
	},


	resample_preview: function() {

		this.commandData.executeCommand("gfx def field " + this.previewOutputFieldName +
			" image_resample field " + this.outputFieldName + 
      " sizes " + this.previewResolution + " " + this.previewResolution + ";");

	},

	apply_filter: function() {
		
		this.outputField.setTypeConnectedThresholdImageFilter(this.inputField,this.belowValue,this.aboveValue,
     this.replaceValue,this.numSeedPoints,this.dimension,this.seedValues,this.numSeedValues);
		this.callback.run();

	},

  preview_filter: function() {

		// update the preview output field options
		this.outputField.setTypeConnectedThresholdImageFilter(this.inputField,this.belowValue,this.aboveValue,
     this.replaceValue,this.numSeedPoints,this.dimension,this.seedValues,this.numSeedValues);

  },


	set_seed: function() {

		// try fetching seed point
		var commandData = this.commandData;
		var parent = this.parent;
		var i;
		var seedNodes = new Object();
		var nodeCount = new Object();

		// pjb: do we need to check the selected nodes are all in the appropriate region?
		// if we define the nodes using preview_coordinates instead of coordinates this
		// may provide a way to check.
		commandData.getSelectedNodes(seedNodes,nodeCount);

		nodeCount = nodeCount.value;

		if(nodeCount == 0) {
			parent.alert("There are no selected nodes!");
		} else {

			//		parent.alert("There are " + nodeCount + " selected nodes");

			// just grab the first node at the moment 		
			var seedNodes = seedNodes.value; 
			
			
			// Currently the interface is only dealing with one node
			// note that the filter itself is capable of dealing with multiple seed
			// nodes, so we should in theory by able to add more seeds.
			// this is an extension that should be added at some point.
			
			var time = 0;
			var coordValues = new Object();
			var coordCount = new Object();
			
			// extract coords
			var coords = commandData.findFieldByName("sliceCoordinates2D");
			coords.evaluateAtNode(seedNodes[0], time, coordValues, coordCount);
			var xcoord = coordValues.value[0]; // grab first value in values array.
			var ycoord = coordValues.value[1]; // grab first value in values array.
						
			var currentNode = null;
			var xcoord = new Array();
			var ycoord = new Array();

			var s = 0;
			this.numSeedValues = 0;

			for (var i = 0; i < nodeCount; i++) {
				
				currentNode = seedNodes[i];
				coords.evaluateAtNode(currentNode, time, coordValues, coordCount);
				this.seedValues[s++] = coordValues.value[0]; // grab first value in values array.
				this.numSeedValues++;

				this.seedValues[s++] = coordValues.value[1]; // grab first value in values array.
				this.numSeedValues++;

			}

			// calculate the number of seed points
			this.numSeedPoints = this.numSeedValues / this.dimension;
			
			// extract intensity value from grayscale field
			
			var compValues = new Object();
			var compCount = new Object();
			var grayscale = commandData.findFieldByName("input_image_nodal_lookup");
			grayscale.evaluateAtNode(seedNodes[0], time, compValues, compCount);
			var intensity = compValues.value[0];
			
			//		parent.alert("The grayscale value at the seed node is " + intensity);
			
			this.belowValue = intensity - 0.1;
			// make sure value is within bounds
			if (this.belowValue < 0 ) {
				this.belowValue = 0;
			}
			
			this.aboveValue = intensity + 0.1;
			// make sure value is within bounds
			if (this.aboveValue > 1) {
				this.aboveValue = 1;
			}
		}
	},


	create_dialog: function() {

		var parent = this.parent;
		var params={connectedThresholdFilter:this};

		this.setup_preview_field();

		try {

			dialogReference = parent.openDialog("chrome://zinc/content/filters/connectedThresholdFilterDialog.xul",
																					"connected-threshold-filter-dialog",
																					"chrome",params);
									
			// store reference to dialog and associated field name
			parent.connectedThresholdFilterDialogReference = dialogReference;
			parent.connectedThresholdFilterDialogReferenceOutputField = this.outputFieldName;

		}catch(ex) {
			parent.alert("Error creating dialog to edit connected threshold filter: " + ex);
		}

	},


  open_dialog: function(gElementName, sceneViewer) {

		this.gElementName = gElementName;
		this.sceneViewer = sceneViewer;

		// parent window
		var parent = this.parent;
		var dialogReference = parent.connectedThresholdFilterDialogReference;
		
		
		// if no reference to an open dialog already exists create it
		if (dialogReference == null || dialogReference.closed) {
			
			// no dialog so create one
			this.create_dialog();
			
		} else {
			
			// dialog exists, check if it is relevant to this field.				
			var oldOutputFieldName = parent.connectedThresholdFilterDialogReferenceOutputField;
			
			if (oldOutputFieldName == this.outputFieldName) {
				
				// correct dialog so bring it to the front
				dialogReference.focus();
				
			} else {
				
				// not correct so close dialog and open a new one
				dialogReference.close();
				this.create_dialog();
				
			}
		}
			
  },


    
  // the QueryInterface function throws an exception
  // if the cmguiIConnectedThresholdFilter interface is not implemented
  // otherwise it returns a reference to the ConnectedThresholdFilter object
  QueryInterface: function(aIID)
  {
    if (!aIID.equals(cmguiIConnectedThresholdFilter) &&    
        !aIID.equals(nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
};

/***********************************************************
class factory

This object is a member of the global-scope Components.classes.
It is keyed off of the contract ID. Eg:

myConnectedThresholdFilter = Components.classes["@bioeng.auckland.ac.nz/helloworld;1"].
                          createInstance(Components.interfaces.cmguiIConnectedThresholdFilter);

***********************************************************/
var ConnectedThresholdFilterFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new ConnectedThresholdFilter()).QueryInterface(aIID);
  }
};

/***********************************************************
module definition (xpcom registration)
***********************************************************/
var ConnectedThresholdFilterModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, 
        CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return ConnectedThresholdFilterFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

/***********************************************************
module initialization

When the application registers the component, this function
is called.
***********************************************************/
function NSGetModule(aCompMgr, aFileSpec) { return ConnectedThresholdFilterModule; }
